###############################################################################
#                 BAYESIAN TREEVARIATE COPULA LEARNING                        #
###############################################################################

## Initialize MCMC algorithm by simple clustering followed by
## maximum likelihood trees

cluster.data <- function(train, K, marg.type, cop.type = "")
{
   if (length(marg.type) == 1)
       marg.type = rep(marg.type, ncol(train))
       
   if (cop.type != "")
      # Expensive!
      return(k.means.treebased(train, K, marg.type, cop.type) - 1)
   else
      # Cheap and dirty
      return(k.means(train, K) - 1)
}

## Calculates the maximum likelihood tree structure by calculating pairwise
## Spearman's rho for each pair and doing greedy search. This also returns
## the rank correlation matrix.

mle.tree.rho <- function(train)
{
   # Initialization
   
   n.vars <- ncol(train)

   if (n.vars == 1)
      return(list(trees = 0, params = 1))
         
   w <- rep(-Inf, n.vars * n.vars)
   corr.m <- rep(1, n.vars * n.vars)
   adj.m <- matrix(rep(0, n.vars * n.vars), ncol = n.vars)

   if (nrow(train) <= 1) {
      # No training data used as a flag to creat a 'default' chain tree
      for (i in 1:(n.vars - 1)) {
          adj.m[i, i + 1] <- 1
          adj.m[i + 1, i] <- 1
      }          
      corr.m <- rep(0, n.vars * n.vars)
      return(list(trees = adj.m, params = corr.m))
   }
      
   # Populate entries with Spearman's rho
   
   for (i in 1:(n.vars - 1)) {
       r.i <- rank(train[, i])
       for (j in (i + 1):n.vars) {
           r.j <- rank(train[, j])
           w[(i - 1) * n.vars + j] <- abs(cor(r.i, r.j))
           corr.m[(i - 1) * n.vars + j] <- cor(r.i, r.j)
           corr.m[(j - 1) * n.vars + i] <- corr.m[(i - 1) * n.vars + j]
           if (is.na(w[(i - 1) * n.vars + j]))
              w[(i - 1) * n.vars + j] <- 0
       }
   }
   
   # Do the greedy search
   
   for (e in 1:(n.vars - 1)) {
       while (T) {
          max.e <- max(w)
          max.e.pos <- which(w == max.e)
          if (length(max.e.pos) > 1) {
             #print(max.e.pos)
             #print(w)
             #print(max.e)
             max.e.pos <- max.e.pos[1]
          }
          e.i <- (max.e.pos - 1) %/% n.vars + 1
          e.j <- max.e.pos - (e.i - 1) * n.vars
          #cat("max.e.pos = ", max.e.pos, " - [", e.i, ", ", e.j, "]", 
          #    ", weight = ", max.e, "\n", sep = "")
          if (is.connected(e.i, e.j, adj.m)) {
             w[max.e.pos] <- -Inf
             next
          }
          adj.m[e.i, e.j] <- 1 
          adj.m[e.j, e.i] <- 1
          w[max.e.pos] <- -Inf
          break
       }
   }
   
   return(list(tree = adj.m, params = corr.m))
}

# Returns true iif v1 and v2 are connected by some path
# in the graph represented by the given adjacency matrix.

is.connected <- function(v1, v2, adj.matrix) {
    
    is.connected.recursion <- function(node) {        
        for (i in 1:n.vars) {        
            if (!adj.matrix[i, node])
                next
            if (i == v2)
                return(TRUE)
            if (i == v1)
                next
            if (marked[i])
                next
            marked[i] <<- TRUE
            result <- is.connected.recursion(i)
            if (result)
                return(TRUE)
        }
        return(FALSE)
    }
    
    n.vars <- nrow(adj.matrix)
    marked <- rep(FALSE, n.vars)
    is.connected.recursion(v1)
}

## Initializes a clustering of size K by k-means

k.means <- function(train, K)
{
    n <- nrow(train)
    p <- ncol(train)
    train <- as.matrix(normalize.data(train))
    cl <- rep(0, n)
    for (i in 1:n)
        cl[i] <- sample(1:K, 1)

    while (T) {
        modified <- F

        centroids <- matrix(rep(0, K * p), nrow = K)
        counts <- rep(0, K)
        for (i in 1:n) {
            k <- cl[i]
            centroids[k, ] <- centroids[k, ] + as.vector(train[i, ])
            counts[k] <- counts[k] + 1;
        }
        for (k in 1:K)
            centroids[k, ] <- centroids[k, ] / counts[k]

        for (i in 1:n) {
            best <- -1
            best.dist <- Inf;

            for (k in 1:K) {
                if (counts[k] == 0)
                    next
                dif <- train[i, ] - centroids[k, ]
                d <- sqrt(sum(dif * dif))
                #cat("d =", d, " best.dist =", best.dist, "\n")
                if (d < best.dist)  {
                    best <- k
                    best.dist <- d
                }
            }

            if (best != cl[i])
                modified <- T

            cl[i] <- best;
        }

        if (!modified) 
            break
    }
    return(cl)
}

k.random <- function(train, K)
{
    n <- nrow(train)
    cl <- rep(0, n)
    for (i in 1:n)
        cl[i] <- sample(1:K, 1)
   return(cl)
}

## A much more expensive k-means algorithm, but that might be more useful.
## Given a cluster assignment, calculate the MLE tree for each group.
## Then compute the log-likelihood of each point with respect to each
## cluster, and iterate. It is not clear if this will always converge,
## but this is not important for initialization purposes.
##
## This assumes a model for the copulas, and that all copulas have the
## same type.
##
## We transform the data to have the empirical CDFs as the marginal models.

k.means.treebased <- function(train, K, marg.type, copula.type)
{
    n <- nrow(train)
    p <- ncol(train)
    train <- marginal.mle.transform(train, marg.type)
    #train <- empirical.cdf(train) #TODO TODO TODO: USE MARG TYPE
    cl <- rep(0, n)
    counts <- rep(0, K)
    for (i in 1:n) {
        cl[i] <- sample(1:K, 1)
        counts[cl[i]] <- counts[cl[i]] + 1
    }
    
    iter.count <- 0
    while (iter.count < 100) {
        cat("K-means initialization, round", iter.count)
        num.modified <- 0

        model <- list()
        for (k in 1:K) {
            if (counts[k] == 0) {
                model[[k]] <- 0
                next                
            }
            train.sel <- train[which(cl == k),,drop = FALSE]
            model[[k]] <- mle.tree.copula(train.sel, copula.type)
        }

        for (i in 1:n) {
            best <- -1
            best.dist <- Inf;

            for (k in 1:K) {
                if (counts[k] == 0)
                    next
                score <- -pred.copula.loglikelihood(model[[k]]$tree, 
                                     model[[k]]$params, train[i, ], copula.type)
                if (score < best.dist)  {
                    best <- k
                    best.dist <- score
                }
            }

            if (best != cl[i])
                num.modified <- num.modified + 1

            cl[i] <- best;           
        }

        cat(" - number of modifications =", num.modified, "\n")
        if (num.modified == 0)  {
            cat("\n")
            break            
        }
        iter.count <- iter.count + 1
    }
    return(cl)
}


##
## Transforms data "dat" by passing it through the corresponding CDFs. The
## CDFs are estimated first. "marg.type" is allowed to be different at every
## variable.
##

marginal.mle.transform <- function(dat, marg.type)
{
   if (marg.type == "" || marg.type == "empirical")
       return(empirical.cdf(dat))

   n <- ncol(dat)
   out.dat <- matrix(0, nrow(dat), ncol(dat))
   for (d in 1:n) {
       if (marg.type[d] == "norm") {
           out.dat[, d] <- F.mle.gamma(dat[, d])$z
       } else if (marg.type[d] == "t") {
           out.dat[, d] <- F.mle.t(dat[, d])$z
       } else if (marg.type[d] == "gamma") {
           out.dat[, d] <- F.mle.gamma(dat[, d])$z
       } else if (marg.type[d] == "exponential") {
           out.dat[, d] <- F.mle.exp(dat[, d])$z
       } else {
           cat("\n")
           stop(paste("\n UNKNOWN MARGINAL TYPE:", marg.type[d], "\n"))
       }
   }
   return(out.dat)
}

## Finds the MLE copula estimator.
##
## IT IS ASSUMED THAT THE DATA HAS BEEN TRANSFORMED BY PASSING IT
## THROUGH THE RESPECTIVE CDFS!
##
## OUTPUT: a list containing the tree and the parameters 
##         (for all pairs, not only the neighbors in the tree.)

mle.tree.copula <- function(train, copula.type)
{
   # Initialization
   
   n.vars <- ncol(train)

   if (n.vars == 1)
      return(list(trees = 0, params = 1))
         
   w <- rep(-Inf, n.vars * n.vars)
   theta.m <- rep(1, n.vars * n.vars)
   adj.m <- matrix(rep(0, n.vars * n.vars), ncol = n.vars)

   if (nrow(train) <= 1) {
      #'Default' chain tree
      for (i in 1:(n.vars - 1)) {
          adj.m[i, i + 1] <- 1
          adj.m[i + 1, i] <- 1
      }          
      theta.m <- rep(0, n.vars * n.vars)
      return(list(trees = adj.m, params = theta.m))
   }
      
   # Populate entries with log-likelihood according to the same type of copula
   
   for (i in 1:(n.vars - 1)) {
       for (j in (i + 1):n.vars) {
           observed.data <- intersect(which(is.na(train[, i]) == F), 
                                      which(is.na(train[, j]) == F))
           if (length(observed.data) <= 2)
               stop("Data contains columns with not enough data!")
           
           cop.output <- fit.copula(train[observed.data, i], 
                                    train[observed.data, j], copula.type)           
                                    
           w[(i - 1) * n.vars + j] <- cop.output$llik
           theta.m[(i - 1) * n.vars + j] <- cop.output$param
           theta.m[(j - 1) * n.vars + i] <- theta.m[(i - 1) * n.vars + j]
       }
   }
   
   # Do the greedy search
   
   for (e in 1:(n.vars - 1)) {
       while (T) {
          max.e <- max(w)
          max.e.pos <- which(w == max.e)
          if (length(max.e.pos) > 1)
             max.e.pos <- max.e.pos[1]
          e.i <- (max.e.pos - 1) %/% n.vars + 1
          e.j <- max.e.pos - (e.i - 1) * n.vars
          #cat("max.e.pos = ", max.e.pos, " - [", e.i, ", ", e.j, "]", 
          #    ", weight = ", max.e, "\n", sep = "")
          if (is.connected(e.i, e.j, adj.m)) {
             w[max.e.pos] <- -Inf
             next
          }
          adj.m[e.i, e.j] <- 1 
          adj.m[e.j, e.i] <- 1
          w[max.e.pos] <- -Inf
          break
       }
   }
   
   return(list(tree = adj.m, params = theta.m))
}

fit.copula <- function(u, v, copula.type)
{
   theta   <- 0
   mle.out <- NULL
   llik    <- -Inf
   if (copula.type == "norm") {
      mle.out <- cop.mle.norm(u, v)
   } else if (copula.type == "fgm") {
      mle.out <- cop.mle.fgm(u, v)
   } else if (copula.type == "amh") {
      mle.out <- cop.mle.amh(u, v)
   } else if (copula.type == "clayton") {
      mle.out <- cop.mle.clayton(u, v)
   } else if (copula.type == "gb") {
      mle.out <- cop.mle.gb(u, v)
   } else if (copula.type == "frank") {
      mle.out <- cop.mle.frank(u, v)
   } else {
      cat("\n")
      stop(paste("\n MLE K-MEANS INITIALIZATION: copula of type", copula.type, 
                 "not yet supported!\n"))
   }
   theta <- mle.out$par
   llik  <- mle.out$llik
   return(list(param = theta, llik = llik))
}

# A function that returns information to be passed to the C program.
#
# This function returns an integer array where each p * p block corresponds
# to an adjacency matrix of a tree. This is done by first selecting the
# data that falls in one of the clusters (cluster indicators are given in
# indices *starting from ZERO*). 
# 

mle.trees.array.C <- function(train, clusters, K, cop.type = "")
{
   p <- ncol(train)
   if (cop.type != "")
       train <- empirical.cdf(train)
   output.trees <- rep(0, K * p * p)
   output.params <- rep(0, K * p * p)
   for (i in 1:K) {
      cl <- i - 1
      sel.rows <- which(clusters == cl)
      cat("Finding initialization tree for cluster #", i, " of size ", 
          length(sel.rows), "\n", sep = "")
      train.sel <- train[sel.rows,,drop = FALSE]
      if (cop.type != "")
          model <- mle.tree.copula(train.sel, cop.type)
      else
          model <- mle.tree.rho(train.sel)
      adj.matrix <- model$tree
      params.matrix <- model$params
      params.matrix[is.na(params.matrix)] <- 0
      start.block <- (i - 1) * p * p + 1
      end.block <- i * p * p
      output.trees[start.block:end.block] <- as.integer(adj.matrix)
      output.params[start.block:end.block] <- as.double(params.matrix)
   }                          
   return(list(trees = output.trees, params = output.params))
}

