###############################################################################
#                 BAYESIAN TREEVARIATE COPULA LEARNING                        #
###############################################################################

## Functions for (quasi-)maximum likelihood (ML) estimation for bivariate joint 
## densities defined by a copula and a pair of marginals. To be more precise,
## such functions implement the estimator where marginals are first first
## individually by MLE, fixed, and then used to generated pseudo-data to fit
## the copula parameters (also by MLE). Functions here are used essentially as
## a way of providing initial parameter positions for the Markov chain.

######################## 
######################## MARGINAL LOG-LIKELIHOODS AND DERIVATIVES
######################## 

################ Gaussian marginals

## F.llik.norm
##
## return the log likelihood of x under a Normal distribution
## with mean mu and variance s2, and the ranks (evaluations
## under the CDF) of x for those parameters -- ro specifies
## Ranks Only.

F.llik.norm <- function(x, mu, s2, ro = FALSE)
{
  if (s2 <= 0) return(list(llik = -Inf))
  if (!ro) llik <- sum(dnorm(x, mean = mu, sd = sqrt(s2), log = TRUE))
  else llik = NA
  z <- pnorm(x, mean = mu, sd = sqrt(s2))
  l.x <- length(z)
  for (i in 1:l.x) {
      if (z[i] == 1)
          z[i] = 0.99999;
      if (z[i] == 0)
          z[i] = 0.00001;
  }
  return(list(llik = llik, z = z))
}

## F.dllik.norm
##
## evaluate the derivative of the log likelihood of a
## normal marginal

F.dllik.norm <- function(x, mu, s2)
{
  n <- length(x)
  dmu <- (sum(x) - n * mu) / s2
  ds2 <- - n / (2 * s2) + sum((x - mu)^2) / (2 * s2^2)
  z <- pnorm(x, mean = mu, sd = sqrt(s2))
  l.x <- length(z)
  for (i in 1:l.x) {
      if (z[i] == 1)
          z[i] = 0.99999;
      if (z[i] == 0)
          z[i] = 0.00001;
  }
  return(list(grad = c(dmu, ds2), z = z))
}

## F.mle.norm:
##
## ML estimation for a normal sample x; returns the ML
## parameterization and the ranks (evaluations under the CDF).

F.mle.norm <- function(x)
{
  m <- mean(x)
  s2 <- mean((x - m)^2)
  z <- pnorm(x, mean=m, sd=sqrt(s2))
  l.x <- length(z)
  for (i in 1:l.x) {
      if (z[i] == 1)
          z[i] = 0.99999;
      if (z[i] == 0)
          z[i] = 0.00001;
  }
      
  llik <- sum(dnorm(x, mean = m, sd = sqrt(s2), log = TRUE))
  return(list(params = data.frame(m = m, s2 = s2), z = z, llik = llik))
}

################ t marginals

## F.llik.exp
##
## return the log likelihood of x under a Student-t distribution
## with position parameter mu and scale s2, and the ranks (evaluations
## under the CDF) of x for those parameters -- ro specifies Ranks Only.

F.llik.exp <- function(x, lambda, ro = FALSE)
{
  if (s2 <= 0 || nu <= 0) return(list(llik = -Inf))
  sig <- sqrt(s2)
  y <- (x - mu)/sig
  if (!ro) llik <- sum(dt(y, df = nu, log = TRUE)) - length(x) * log(sig)
  else llik = NA
  z <- pt(y, df = nu)
  return(list(llik = llik, z = z))
}

# Similar, but more appropriate for optim

F.llik.t2 <- function(params, x, nu)
{
  mu <- params[1]
  s2 <- params[2]

  if (s2 <= 0) {
      return(-Inf)
  }

  sig <- sqrt(s2)
  y <- (x - mu)/sig

  return(sum(dt(y, df = nu, log = TRUE)) - length(x) * log(sig))
}

## F.dllik.t
##
## evaluate the derivative of the log likelihood of a
## student-t marginal

F.dllik.t <- function(x, nu, mu, s2)
{
  n <- length(x)

  ## nu part of the gradient
  dnu <- n * (digamma((nu + 1) / 2) - digamma(nu / 2) - 1 / nu)
  dnu <- dnu - sum(log(1 + (x - mu)^2 / (nu * s2)))
  dnu <- dnu + (nu + 1) * sum((x - mu)^2 / (nu^2 * s2 + nu * (x - mu)^2))
  dnu <- 0.5 * dnu

  ## mu part of the gradient
  dmu <- (nu + 1) * sum((x - mu ) /(nu * s2 + (x - mu)^2))

  ## s2 part of the gradient
  ds2 <- -n / s2 + (nu + 1) * sum((x - mu)^2 / (nu * s2^2 + (x - mu)^2 * s2))
  ds2 <- 0.5 * ds2

  ## create the ranks
  y <- (x - mu) / sqrt(s2)
  z <- pt(y, df = nu)

  return(list(grad = c(dnu, dmu, ds2), z = z))
}

# Similar, but more appropriate for optim

F.dllik.t2 <- function(params, x, nu)
{
  n <- length(x)
  mu <- params[1]
  s2 <- params[2]
  
  ## mu part of the gradient
  dmu <- (nu + 1) * sum((x - mu ) /(nu * s2 + (x - mu)^2))

  ## s2 part of the gradient
  ds2 <- -n / s2 + (nu + 1) * sum((x - mu)^2 / (nu * s2^2 + (x - mu)^2 * s2))
  ds2 <- 0.5 * ds2

  return(c(dmu, ds2))
}

## F.mle.t:
##
## "ML" estimation for a normal sample x; returns the ML
## parameterization and the ranks (evaluations under the CDF).
## Notice that degrees of freedom "mu" are fixed to 5 instead of estimated.

F.mle.t <- function(x)
{
  nu <- 5 # Fixed degrees of freedom
  params <- optim(par = c(0, 1), fn = F.llik.t2, gr = F.dllik.t2,
             method = "BFGS", #lower = -Inf, upper = Inf,
             control = list(fnscale = -1), x = x, nu = nu)

  mu <- params$par[1]
  s2 <- params$par[2]
  
  y <- (x - mu) / sqrt(s2)
  z <- pt(y, df = nu)
  l.x <- length(z)
  for (i in 1:l.x) {
      if (z[i] == 1)
          z[i] = 0.99999;
      if (z[i] == 0)
          z[i] = 0.00001;
  }
      
  llik <- sum(dt(y, df = nu, log = TRUE))
  return(list(params = data.frame(mu = mu, nu = nu, s2 = s2), z = z, llik = llik))
}

################ Exponential marginals

## F.llik.exp
##
## return the log likelihood of x under a Student-t distribution
## with position parameter mu and scale s2, and the ranks (evaluations
## under the CDF) of x for those parameters -- ro specifies Ranks Only.

F.llik.exp <- function(x, lambda, ro = FALSE)
{
  if (!ro) llik <- sum(dexp(y, rate = lambda, log = TRUE))
  else llik = NA
  z <- pexp(x, rate = lambda)
  return(list(llik = llik, z = z))
}

## F.dllik.exp
##
## evaluate the derivative of the log likelihood of a
## exponential marginal.

F.dllik.exp <- function(x, lambda)
{
  n <- length(x)

  ## gradient with respect to lambda
  dlambda <- n / lambda - sum(x)
  
  ## create the ranks
  z <- pexp(x, rate = lambda)

  return(list(grad = c(lambda), z = z))
}

## F.mle.exp:
##
## "ML" estimation for an exponential sample x; returns the ML
## parameterization and the ranks (evaluations under the CDF).

F.mle.exp <- function(x)
{
  lambda <- 1 / mean(x)
  z <- pexp(x, rate = lambda)
  l.x <- length(z)
  for (i in 1:l.x) {
      if (z[i] == 1)
          z[i] = 0.99999;
      if (z[i] == 0)
          z[i] = 0.00001;
  }
      
  llik <- sum(dexp(x, rate = nu, log = TRUE))
  return(list(params = data.frame(lambda = lambda), z = z, llik = llik))
}

################ gamma marginals

## F.llik.gamma
##
## return the log likelihood of x under a gamma distribution
## with shape parameter "a" and scale "s", and the ranks (evaluations
## under the CDF) of x for those parameters -- ro specifies Ranks Only.

F.llik.gamma <- function(x, a, s, ro = FALSE)
{
  if (!ro) llik <- sum(dgamma(x, shape = a, scale = s, log = TRUE))
  else llik = NA
  z <- pgamma(x, shape = a, scale = s)
  return(list(llik = llik, z = z))
}

# Similar, but more appropriate for optim

F.llik.gamma2 <- function(params, x)
{
  if (params[1] <= 0 || params[2] <= 0) {
      return(-Inf)
  }
  return(sum(dgamma(x, shape = params[1], scale = params[2], log = TRUE)))
}

## F.dllik.gamma
##
## evaluate the derivative of the log likelihood of a
## gamma marginal

F.dllik.gamma <- function(x, a, s)
{
  n <- length(x)

  ## shape part of the gradient
  da <- sum(log(x)) - n * (digamma(a) + log(s))

  ## scale part of the gradient
  ds <- sum(x) / s^2 - n * a / s

  ## create the ranks
  z <- pgamma(x, shape = a, scale = s)

  return(list(grad = c(da, ds), z = z))
}

# Similar, but more appropriate for optim

F.dllik.gamma2 <- function(params, x)
{
  n <- length(x)
  a <- params[1]
  s <- params[2]

  ## shape part of the gradient
  da <- sum(log(x)) - n * (digamma(a) + log(s))

  ## scale part of the gradient
  ds <- sum(x) / s^2 - n * a / s

  return(c(da, ds))
}

## F.mle.gamma:
##
## "ML" estimation for a gamma sample x; returns the ML
## parameterization and the ranks (evaluations under the CDF).

F.mle.gamma <- function(x)
{
  params <- optim(par = c(1, 1), fn = F.llik.gamma2, gr = F.dllik.gamma2,
             method = "BFGS", #lower = c(0, 0), upper = c(Inf, Inf),
             control = list(fnscale = -1), x = x)

  a <- params$par[1]
  s <- params$par[2]
  
  z <- pgamma(x, shape = a, scale = s)
  l.x <- length(z)
  for (i in 1:l.x) {
      if (z[i] == 1)
          z[i] = 0.99999;
      if (z[i] == 0)
          z[i] = 0.00001;
  }
      
  llik <- sum(dgamma(x, shape = a, scale = s, log = TRUE))
  return(list(params = data.frame(shape = a, scale = s), z = z, llik = llik))
}

######################## 
######################## COPULA LOG-LIKELIHOODS AND DERIVATIVES
######################## 

## All inputs 'u' and 'v' are assumed to be uniform (0, 1) variables.

################ Gaussian copula

## cop.llik.norm:
##
## log likelihood of a normal (Gaussian) copula parameterized by theta.

cop.llik.norm <- function(theta, u, v)
{
  ## calculate sufficient statistics
  N <- length(u)
  A <- sum(qnorm(u)^2)
  B <- sum(qnorm(v)^2)
  C <- sum(qnorm(u) * qnorm(v))

  ## evaluate likelihood
  if(theta <= -1 || theta >= 1) return(-Inf)
  ll <- -(N / 2) * log(1 - theta^2)
  ll <- ll - (theta^2 * (A + B) - 2 * theta * C)/(2 * (1 - theta^2))
  return(ll)
}

## cop.dllik.norm:
##
## derivative of log likelihood of a normal (Gaussian) copula
## parameterized by theta, where
## A = sum(qnorm(u)^2)
## B = sum(qnorm(v)^2)
## C = sum(qnorm(u)*qnorm(v))

cop.dllik.norm <- function(theta, u, v)
{
  ## calculate sufficient statistics
  N <- length(u)
  A <- sum(qnorm(u)^2)
  B <- sum(qnorm(v)^2)
  C <- sum(qnorm(u)*qnorm(v))
  
  ## evaluate derivative
  if(theta <= -1 || theta >= 1) return(-Inf)
  dll <- theta * (1 - theta^2) * N
  dll <- dll - (theta * A + theta * B - C - theta^2 * C)
  dll <- dll / (1 - theta^2)^2
  return(dll)
}


## cop.mle.norm:
##
## finds the ML estimate of the parameter theta to a normal
## copula using a gradient based optimization method.

cop.mle.norm <- function(u, v)
{
  o <- optim(par = 0, fn = cop.llik.norm, gr = cop.dllik.norm,
             #method="L-BFGS-B", lower=-0.999, upper=0.999,
             method = "BFGS", lower = -Inf, upper = Inf,
             control = list(fnscale = -1), u = u, v = v)

  return(list(par = o$par, llik = o$value))
}

################ Frank copula

## cop.llik.frank:
##
## log likelihood of a Frank copula parameterized by theta.

cop.llik.frank <- function(theta, u, v)
{
  if (theta == 0) {
      return(-Inf)
  }
  e.uv <- exp(-theta * (u + v))
  e.t  <- exp(-theta)
  e.u  <- exp(-theta * u)
  e.v  <- exp(-theta * v)
  return (sum(log(-theta * 
                    (e.uv * (e.t - 1)) / (e.t + e.uv - e.u - e.v)^2)))
}

## cop.dllik.frank:
##
## derivative of log likelihood of a Frank copula parameterized by theta.

cop.dllik.frank <- function(theta, u, v)
{
  e.uv <- exp(-theta * (u + v))
  e.t  <- exp(-theta)
  e.u  <- exp(-theta * u)
  e.v  <- exp(-theta * v)
  
  return(sum(1 / theta + e.t / (1 - e.t) - (u + v) + 
             2 * (e.t + (u + v) * e.uv - u * e.u - v * e.v) / 
                 (e.t + e.uv - e.u - e.v)))
}


## cop.mle.frank:
##
## finds the ML estimate of the parameter theta to a Frank
## copula using a gradient based optimization method.

cop.mle.frank <- function(u, v)
{
  o <- optim(par = 1, fn = cop.llik.frank, gr = cop.dllik.frank,
             #method="L-BFGS-B", lower=-0.999, upper=0.999,
             method = "BFGS", lower = -Inf, upper = Inf,
             control = list(fnscale = -1), u = u, v = v)

  return(list(par = o$par, llik = o$value))
}

################ Gumbel-Barnett copula

## cop.llik.gb:
##
## log likelihood of a Gumbel-Barnett copula parameterized by theta.

cop.llik.gb <- function(theta, u, v)
{
  if (theta < 0 || theta > 1) {
      return(-Inf)
  }
  
  ln.u <- log(u)
  ln.v <- log(v)
     
  return (sum(log((1 - theta * ln.u) * (1 - theta * ln.v) - theta) 
               - theta * ln.u * ln.v))
}

## cop.dllik.gb:
##
## derivative of log likelihood of a Gumbel-Barnett copula parameterized by theta.

cop.dllik.gb <- function(theta, u, v)
{
  ln.u <- log(u)
  ln.v <- log(v)
  
  return(sum((2 * theta * ln.u * ln.v - ln.u - ln.v - 1) /
             ((1 - theta * ln.u) * (1 - theta * ln.v) - theta)))
}


## cop.mle.gb:
##
## finds the ML estimate of the parameter theta to a Gumbel-Barnett
## copula using a gradient based optimization method.

cop.mle.gb <- function(u, v)
{
  o <- optim(par = 0, fn = cop.llik.gb, gr = cop.dllik.gb,
             #method="L-BFGS-B", lower=-0.999, upper=0.999,
             method = "BFGS", lower = -Inf, upper = Inf,
             control = list(fnscale = -1), u = u, v = v)

  return(list(par = o$par, llik = o$value))
}

################ Ali-Mikhail-Haq copula

## cop.llik.amh:
##
## log likelihood of a Ali-Mikhail-Haq copula parameterized by theta.

cop.llik.amh <- function(theta, u, v)
{
  if (theta < -1 || theta > 1) {
      return(-Inf)
  }
  
  return (sum(log(1 - theta + 2 * theta * u * v - theta * (1 - u) * (1 - v) + (1 - u) * (1 - v) * theta^2) - 
              3 * log(1 - theta * (1 - u) * (1 - v))))
}

## cop.dllik.amh:
##
## derivative of log likelihood of a Ali-Mikhail-Haq copula parameterized by theta.

cop.dllik.amh <- function(theta, u, v)
{
  one.uv <- (1 - u) * (1 - v)
  return(sum((-1 + 2 * u * v + (2 * theta - 1) * one.uv) / 
             (1 - theta + 2 * theta * u * v - theta * one.uv + one.uv * theta^2)
                 +
             (3 * one.uv) / (1 - theta * one.uv)))
}


## cop.mle.amh:
##
## finds the ML estimate of the parameter theta to a Ali-Mikhail-Haq
## copula using a gradient based optimization method.

cop.mle.amh <- function(u, v)
{
  o <- optim(par = 0, fn = cop.llik.amh, gr = cop.dllik.amh,
             #method="L-BFGS-B", lower=-0.999, upper=0.999,
             method = "BFGS", lower = -Inf, upper = Inf,
             control = list(fnscale = -1), u = u, v = v)

  return(list(par = o$par, llik = o$value))
}

################ Clayton copula

## cop.llik.clayton:
##
## log likelihood of a Clayton copula parameterized by theta.

cop.llik.clayton <- function(theta, u, v)
{
  if (theta < -1 || theta == 0) {
      return(-Inf)
  }

  return (sum(log(theta + 1) - (theta + 1) * log(u * v) -
              (2 * theta + 1) / theta * log(u^(-theta) + v^(-theta) - 1)))
}

## cop.dllik.clayton:
##
## derivative of log likelihood of a Clayton copula parameterized by theta.

cop.dllik.clayton <- function(theta, u, v)
{
  u.t <- u^(-theta)
  v.t <- v^(-theta)
  return(sum(1 / (theta + 1) - log(u * v) +
             1 / theta^2 * log(u.t * v.t - 1) +
             (2 + 1 / theta) * (u.t * log(u) + v.t * log(v)) * (u.t + v.t - 1)))
}


## cop.mle.clayton:
##
## finds the ML estimate of the parameter theta to a Clayton
## copula using a gradient based optimization method.

cop.mle.clayton <- function(u, v)
{
  o <- optim(par = 1, fn = cop.llik.clayton, gr = cop.dllik.clayton,
             #method="L-BFGS-B", lower=-0.999, upper=0.999,
             method = "BFGS", lower = -Inf, upper = Inf,
             control = list(fnscale = -1), u = u, v = v)

  return(list(par = o$par, llik = o$value))
}

################ Farlie-Gumbel-Morgenstern copula

## cop.llik.fgm:
##
## log likelihood of a Farlie-Gumbel-Morgenstern copula parameterized by theta.

cop.llik.fgm <- function(theta, u, v)
{
  if (theta < -1 || theta > 1) {
      return(-Inf)
  }
  
  return (sum(log(1 + theta * (1 - 2 * u) * (1 - 2 * v))))
}

## cop.dllik.fgm:
##
## derivative of log likelihood of a Farlie-Gumbel-Morgenstern copula 
## parameterized by theta.

cop.dllik.fgm <- function(theta, u, v)
{
  one.uv <- (1 - 2 * u) * (1 - 2 * v)
  return(sum((one.uv / (1 + theta * one.uv))))
}


## cop.mle.fgm:
##
## finds the ML estimate of the parameter theta to a Farlie-Gumbel-Morgenstern
## copula using a gradient based optimization method.

cop.mle.fgm <- function(u, v)
{
  o <- optim(par = 0, fn = cop.llik.fgm, gr = cop.dllik.fgm,
             #method="L-BFGS-B", lower=-0.999, upper=0.999,
             method = "BFGS", lower = -Inf, upper = Inf,
             control = list(fnscale = -1), u = u, v = v)

  return(list(par = o$par, llik = o$value))
}

######################## 
######################## GENERAL "MLE" FUNCTIONS
######################## 

## cop.all.mle:
##
## sequential estimation of the parameters to the two marginals
## (Fx and Fy) and then estimation of the copula (cop) conditional
## on the resulting ranks under the marginals

cop.all.mle <- function(x, y = NULL, cop.mle = cop.mle.norm,
                         Fx.mle = F.mle.norm, Fy.mle = F.mle.norm)
{
  ## get x and y from the inputs
  if (is.null(y)) {
      y <- x$y
      x <- x$x
  }

  ## get the MLE and quantiles
  xmle <- Fx.mle(x)
  ymle <- Fy.mle(y)

  ## now use them to find the MLE of the copula
  o <- cop.mle(xmle$z, ymle$z)

  ## calculate the actual log likelihood obtained
  llik <- o$llik + xmle$llik + ymle$llik
  
  return(c(list(llik = llik, theta = o$par),
           list(param = data.frame(x = xmle$param, y = ymle$param))))
}

