###############################################################################
#                 BAYESIAN TREEVARIATE COPULA LEARNING                        #
###############################################################################

###### Generic benchmark script

# This file provides a general template for running benchmarks with the XCOP 
# software. In a nutshell, it loads data, removes discrete variables (as an 
# example of pre-processing), sets a series of simulations with particular
# marginals and different samplers. The goal is to generate different trace
# files in order to evaluate the mixing quality of the different proposals,
# similar to the paper
#
# R. Silva and R. Gramacy (2009). "MCMC methods for Bayesian mixtures of
# copulas". Proceedings of Twelfth International Conference on Artificial 
# Intelligence and Statistics. Clearwater, FL.

##
## Runs methods and evaluates results
##

simulation.size <- 1000  # Number of MCMC iterations
burn.in         <- 100   # Initial samples to ignore when doing evaluation
mcmc.step       <- 1     # Dump sample to file only at every "mcmc.step".
                         #  Set this to > 1 to save disk space while discarding
                         #  correlated samples.


train.x <- read.table("../data/finance.train", header = T)
train.x <- train.x[1:200, ] # Just to speed up demonstration
test.x  <- read.table("../data/finance.test" , header = T)

# The following imputes any missing data with a very simple technique, followed
# by normalization. Please notice that in this benchmark we account for missing 
# data too, but in a synthetic, controlled way (see below).

train.x <- complete.set(train.x)
test.x  <- complete.set(test.x)
n.data  <- normalize.datas(train.x, test.x)
train.x <- n.data$train
test.x  <- n.data$test

# A final amount of pre-processing. The sampling procedure requires the data to
# be a specific format (see "stack.data" for details, if desired).

stacked.data.train <- stack.data(train.x)
stacked.data.test  <- stack.data(test.x)
n.vars <- ncol(train.x)

# Set marginals and copulas. For benchmarking, "norm", i.e., the Gaussian 
# copula, is the only possible choice: we have not implemented yet different
# "advanced" samplers for non-Gaussian copulas. See "runme.fit.R" to a generic
# choice of model without concerns on comparing the "smart" samplers to
# standard ones.
#
# The different choice of marginals are: "norm", "t", "gamma", "exponential"
# and "empirical".
#
# If "empirical" is chosen, the sampler will automatically know that marginal
# parameters are fixed. Notice that marginal families can be different for 
# different. The ordering within the "marginals" vector corresponds to the 
# ordering of variables in the datasets.

marginals   <- "empirical"
marg.priors <- list(list())
marg.mh     <- list(list())

copula.type <- "norm"
cop.priors  <- list(list())
cop.mh      <- list(win = 0.15)

# Setting missingness pattern. "p.missing" indicates the probability of marking
# a particular entry in the training data as missing. Set to zero to avoid any
# missing entries. A matrix of indicators "missing.pattern", of the same size
# as the training set, is generated.

p.missing <- 0.0
missing.pattern <- generate.random.missing.pattern(train.x, p.missing)

# Initialize chain: first, do a "model-free" cluster of the data, and within
# each cluster find the maximum likelihood tree in order to set the initial
# tree and parameters. The clustering procedure might be quite slow.

initial.K <- 2
initial.clustering <- cluster.data(train.x, initial.K, copula.type)
initial.models <- mle.trees.array.C(train.x, initial.clustering, 
                                    initial.K, copula.type)
initial.trees  <- initial.models$trees
initial.params <- initial.models$params

# Call algorithm with different tree proposals. Notice we keep track of wallclock
# time as part of the benchmark. "mcmc.options" is a vector of 7 integers 
# indicating several options for the sample as follows:
# 
# Entry 1: when set to 1, it indicates we also want to save in a log file all 
#          missing value imputations during sampling;
# Entry 2: when set to 1, it indicates not to allocate memory for the 
#          predictive log-likelihood, but do it later by reading from the log 
#          files. Reading from log files might be much slower than storing in 
#          memory, but if you don't have enough memory, this will need to be set
#          to 1 if predictive log-likelihoods are to be calculated.
# Entry 3: if this iset to 1 and the Gaussian copula is used for all pairs,
#          then save also the implied copula correlation matrices while 
#          sampling;
# Entry 4: set choice of tree sampler. Set it as follows.
#            0: "treeangle." Choose a triplet X - Y - Z to be modified into
#               Y - Z - X, proposed copula parameter based on current Spearman
#               correlation matrix (X, Y, Z)
#            1: "simple." Choose an edge to be removed uniformly at random,
#               choose a new legal edge to add uniformly at random, propose
#               new copula parameter from prior;
#            2: "hybrid." Similar to "treeangle", but first choose between trees
#                Y - Z - X and Z - X - Y by (numerically) marginalizing 
#               parameters (new copula parameter is proposed just as in 
#               "treeangle").
#            3: "exact."  For a fixed set of parameters, sample a new tree from
#               the exact conditional distribution. Numerically unstable, and
#               not necessarily provide good mixing.
#            PLEASE NOTICE THAT CURRENTLY, FOR COPULAS OTHER THAN THE
#            GAUSSIAN, ONLY THE "SIMPLE" PROPOSAL HAS BEEN IMPLEMENTED. IF YOU
#            CHOOSE A NON-"SIMPLE" PROPOSAL WITH A NON-GAUSSIAN COPULA, IT WILL
#            AUTOMATICALLY CHANGE TO "SIMPLE" BEHIND THE SCENES. WE DO HAVE
#            PLANS TO EXPAND THE PROPOSAL IMPLEMENTATION TO OTHER COPULAS.
# Entry 5: when set to 1, treats marginals as constants (done automatically if
#          marginals are set to "empirical");
# Entry 6: when set to 1, sample copulas from slice sampler instead of random
#          walk Metropolis-Hastings (note: only implemented for Gaussian copula
#          so far. Ignored otherwise).
# Entry 7: when set to 1, explicit sample copula parameters even for those 
#          copulas not associated with any tree. This is necessary only if 
#          one is interested in examining the log files of the cluster with
#          the largest number of associated points.

# Call algorithm with "simple" tree proposal

t1.start <- Sys.time()
mcmc.options <- c(1, 0, 1, 1, 1, 1, 1) 
r1 <- simple.mcmc.treevariate(TRUE, 
        marginals, marg.priors, marg.mh, copula.type, cop.priors, cop.mh,
        simulation.size, burn.in, mcmc.step, 
        ncol(train.x), nrow(train.x), stacked.data.train, 
        missing.pattern,
        nrow(test.x), stacked.data.test, recode.missing.pattern(test.x),
        0.1, 1, initial.K, initial.clustering, initial.trees, initial.params,
        mcmc.options, train.x)
acpt.trees1 <- read.table("accept_tree.mcmc") 
predictive1 <- read.table("predictive.mcmc")
predictive1 <- predictive1[2:nrow(predictive1), ]
implied1 <- read.table("param.impl.dat", header = T)
copula1 <- read.table("copula.sel.dat", header = T)
t1.end <- Sys.time()

#Call algorithm with "treeangle" tree proposal
t2.start <- Sys.time()
mcmc.options <- c(1, 0, 1, 0, 1, 1, 1)
r2 <- simple.mcmc.treevariate(TRUE,
        marginals, marg.priors, marg.mh, copula.type, cop.priors, cop.mh,
        simulation.size, burn.in, mcmc.step, 
        ncol(train.x), nrow(train.x), stacked.data.train, 
        missing.pattern,
        nrow(test.x), stacked.data.test, recode.missing.pattern(test.x),
        0.1, 1, initial.K, initial.clustering, initial.trees, initial.params,
        mcmc.options, train.x)
acpt.trees2 <- read.table("accept_tree.mcmc") 
predictive2 <- read.table("predictive.mcmc")
predictive2 <- predictive2[2:nrow(predictive2), ]
implied2 <- read.table("param.impl.dat", header = T)
copula2 <- read.table("copula.sel.dat", header = T)
t2.end <- Sys.time()

#Call algorithm with "hybrid" tree proposal
t3.start <- Sys.time()
mcmc.options <- c(1, 0, 1, 2, 1, 1, 1)
r3 <- simple.mcmc.treevariate(TRUE,
        marginals, marg.priors, marg.mh, copula.type, cop.priors, cop.mh,
        simulation.size, burn.in, mcmc.step, 
        ncol(train.x), nrow(train.x), stacked.data.train, 
        missing.pattern,
        nrow(test.x), stacked.data.test, recode.missing.pattern(test.x),
        0.1, 1, initial.K, initial.clustering, initial.trees, initial.params,
        mcmc.options, train.x)
acpt.trees3 <- read.table("accept_tree.mcmc") 
predictive3 <- read.table("predictive.mcmc")
predictive3 <- predictive3[2:nrow(predictive3), ]
implied3 <- read.table("param.impl.dat", header = T)
copula3 <- read.table("copula.sel.dat", header = T)
t3.end <- Sys.time()

#Call algorithm with "exact" tree proposal
t4.start <- Sys.time()
mcmc.options <- c(1, 0, 1, 3, 1, 1, 1)
r4 <- simple.mcmc.treevariate(TRUE,
        marginals, marg.priors, marg.mh, copula.type, cop.priors, cop.mh,
        simulation.size, burn.in, mcmc.step, 
        ncol(train.x), nrow(train.x), stacked.data.train, 
        missing.pattern,
        nrow(test.x), stacked.data.test, recode.missing.pattern(test.x),
        0.1, 1, initial.K, initial.clustering, initial.trees, initial.params,
        mcmc.options, train.x)
acpt.trees4 <- read.table("accept_tree.mcmc") 
predictive4 <- read.table("predictive.mcmc")
predictive4 <- predictive4[2:nrow(predictive4), ]
implied4 <- read.table("param.impl.dat", header = T)
copula4 <- read.table("copula.sel.dat", header = T)
t4.end <- Sys.time()

# Display progress

acpt.nrow <- nrow(acpt.trees1)
plot(1:acpt.nrow, acpt.trees1[1:acpt.nrow,1], "l", col="red", 
     main = "Tree acceptance", ylab = "Acceptance", xlab = "Iteration", 
     ylim = c(0, 0.4))
lines(1:acpt.nrow, acpt.trees2[1:acpt.nrow,1], "l", col="blue")
lines(1:acpt.nrow, acpt.trees3[1:acpt.nrow,1], "l", col="green")
lines(1:acpt.nrow, acpt.trees4[1:acpt.nrow,1], "l", col="black")
x11()
min.p <- min(c(predictive1, predictive2, predictive3))
max.p <- max(c(predictive1, predictive2, predictive3))
p.nrow <- length(predictive1)
plot(1:p.nrow, predictive1, "l", col="red", main = "Predictive log-likelihood", 
     ylab = "Average log-likelihood", xlab = "Iteration", 
     ylim = c(min.p - 0.1, max.p + 0.1))
lines(1:p.nrow, predictive2, "l", col="blue")
lines(1:p.nrow, predictive3, "l", col="green")
lines(1:p.nrow, predictive4, "l", col="black")

#Provide comparisons

burn.in <- trunc(burn.in / mcmc.step) # Adjustment, since recorded traces 
                                      # are thinned
simulation.size <- trunc(simulation.size / mcmc.step)

cat("\nSUMMARY OF RESULTS\n")
cat("------------------\n\n")
cat("Predictive log-likelihood (standard structure sampler):", r1[1], "\n")
imp1 <- effectiveSize(implied1[(burn.in + 1):simulation.size, ])
cat("Tree acceptance rate:", r1[3], "\n");
cat("---\n")
cat("Predictive log-likelihood (treevariate structure sampler):", r2[1], "\n")
imp2 <- effectiveSize(implied2[(burn.in + 1):simulation.size, ])
cat("Tree acceptance rate:", r2[3], "\n");
cat("---\n")
cat("Predictive log-likelihood (hybrid structure sampler):", r3[1], "\n")
imp3 <- effectiveSize(implied3[(burn.in + 1):simulation.size, ])
cat("Tree acceptance rate:", r3[3], "\n");
cat("---\n")
cat("Predictive log-likelihood (exact structure sampler):", r4[1], "\n")
imp4 <- effectiveSize(implied4[(burn.in + 1):simulation.size, ])
cat("Tree acceptance rate:", r4[3], "\n");
cat("---\n")

windows()
diff.max <- max(abs(imp2 - imp1))
plot(imp1, imp2, main = "Comparison of average effective sample size",
     xlab = "Standard", ylab = "Treevariate", xlim = c(0, diff.max + 1),
     ylim = c(0, diff.max + 1))
abline(0,1)
cat("Percentage of successul improvements: ", sum(imp1 < imp2) / length(imp1), "\n")

windows()
diff.max <- max(abs(imp3 - imp1))
plot(imp1, imp3, main = "Comparison of average effective sample size",
     xlab = "Standard", ylab = "Hybrid", xlim = c(0, diff.max + 1),
     ylim = c(0, diff.max + 1))
abline(0,1)
cat("Percentage of successul improvements: ", sum(imp1 < imp3) / length(imp1), "\n")

windows()
diff.max <- max(abs(imp3 - imp2))
plot(imp2, imp3, main = "Comparison of average effective sample size",
     xlab = "Treevariate", ylab = "Hybrid", xlim = c(0, diff.max + 1),
     ylim = c(0, diff.max + 1))
abline(0,1)
cat("Percentage of successul improvements: ", sum(imp2 < imp3) / length(imp2), "\n")

windows()
diff.max <- max(abs(imp4 - imp2))
plot(imp4, imp3, main = "Comparison of average effective sample size",
     xlab = "Exact", ylab = "Hybrid", xlim = c(0, diff.max + 1),
     ylim = c(0, diff.max + 1))
abline(0,1)
cat("Percentage of successul improvements: ", sum(imp4 < imp3) / length(imp4), "\n")

cat("Algorithm 1:", t1.end - t1.start, "\n")
cat("Algorithm 2:", t2.end - t2.start, "\n")
cat("Algorithm 3:", t3.end - t3.start, "\n")
cat("Algorithm 4:", t4.end - t4.start, "\n")

###############################################################################
# The output generated by the above call consists only of text files:
#
# - z.dat: sampled values for cluster assignments;
# - trees.dat: sampled values for the trees. It comes in the format
#              "number of clusters" "number of edges tree #1" 
#              (pairs of vertex numbers) "number of edges tree #2", etc.
# - param.dat: parameters for marginals, followed by parameters for copulas.
#              If an edge is not associated with any tree in the mixture for a 
#              particular sample, the value recorded is not NA, but some
#              arbitrary value (makes reading from the file a bit faster - one
#              can always use the tree information to know which parametes are
#              not being used);
# - marginal.dat: redundant file with parameters for marginals only;
# - k.dat: records the number of trees and respective alpha parameter for
#          the Dirichlet process sampled at each step;
# - param.impl.dat: implied correlation matrix (only if all copulas are Gaussian,
#                   and mcmc.options[3] == 1);
