###############################################################################
#                 BAYESIAN TREEVARIATE COPULA LEARNING                        #
###############################################################################

###### Generic running script

# An example script for the main function "copula.tree.mcmc" which implements
# the method described by
#
# R. Silva and R. Gramacy (2009). "MCMC methods for Bayesian mixtures of
# copulas". Proceedings of Twelfth International Conference on Artificial
# Intelligence and Statistics. Clearwater, FL.
#
# Very similar to runme.example1.R. The main difference is as follows.
#
# This example file also randomly pretends some of the entries in the data
# are missing, so that it can compare the performance of the MCMC method
# against a trivial imputation method for filling in missing data. To remove
# any artificial missing data evaluation, just set "prob.miss" to zero in the
# appropriate place:
#
#     prob.miss <- 0
#
# Also, we complement it with simple checks, such as the check for variables
# without enough values.

#######################################################################
## Setup sampler
##

# Number of MCMC iterations
simulation.size <- 1000    

# Reads the data, and discard any columns without at least three different
# values. All data is assumed to be continuous.

train.x <- read.table("../data/finance.train", header = T)
c.filter <- discreteness.filter(train.x, 3)
if (ncol(train.x) - length(c.filter) > 0)
    cat("Removing", ncol(train.x) - length(c.filter),
        "variables due to discreteness!\n")
train.x <- train.x[, c.filter]
test.x  <- read.table("../data/finance.test" , header = T)
test.x  <- test.x[, c.filter]

# The following imputes any test missing data with a very simple technique, 
# followed by normalization. 

test.x  <- complete.set(test.x)
n.data  <- normalize.datas(train.x, test.x)
train.x <- n.data$train
test.x  <- n.data$test

# Set marginals and copulas. The different choice of marginals are: "norm", "t", 
# "gamma", "exponential" and "empirical".
#
# If "empirical" is chosen, the sampler will automatically know that marginal
# parameters are fixed. Notice that marginal families can be different for
# different. The ordering within the "marginals" vector corresponds to the
# ordering of variables in the datasets.

marg.type   <- "t"
marg.priors <- list(list())
marg.mh     <- list(list(mu.s2q = 0.05, s2.prop = 0.70, nu.prop = 0.70))

cop.type    <- "norm"
cop.priors  <- list(list())
cop.mh      <- list(win = 0.15)

mix.priors   <- list()
mix.priors$a <- 0.1
mix.priors$b <- 1

# Initializate chain: first, do a "model-free" cluster of the data.

opt <- list()
opt$initial.K <- 2                 # Number of initial mixture components
opt$init.cluster.t <- "euclidean"  # Criterion for initializing components
opt$fix.marginals <- FALSE         # Marginals treated by Bayesian inference
                                   #    or by fixing them via MLE?
opt$predict.from.file <- FALSE     # If you want to evaluate predictive 
                                   #    log-likelihood in a test set but don't
                                   #    have enough memory to store samples in
                                   #    RAM, set this to TRUE
opt$burn.in <- 1                   # Initial samples to ignore when doing 
                                   #    evaluation.
opt$save.missing <- TRUE           # Save predictions for missing values
prob.miss <- 0.                    # Generate "artificial" missing data for
                                   #    comparison purposes.
opt$missing.pattern <- matrix(generate.random.missing.pattern(train.x, prob.miss), 
                              ncol = ncol(train.x))

# Finally, call procedure and time it

t.start <- Sys.time()
tree.output <- copula.tree.mcmc(simulation.size, train.x, test.x, 
                                marg.type,  marg.priors, marg.mh,
                                cop.type,   cop.priors,  cop.mh,
                                mix.priors, opt)
t.end <- Sys.time()

#######################################################################
## Analyze output
##

## Read output files

# Percentage of accepted tree moves across MCMC evolution
acpt.trees <- read.table("accept_tree.mcmc")
# Predictive log-likelihood on test set
predictive <- read.table("predictive.mcmc")
predictive <- predictive[2:nrow(predictive), ]
# All pairwise copula parameters from the most probable tree
# (this should vary a lot, since parameters for edges not in the tree are
#  sampled from the prior)
copula.log <- read.table("copula.sel.dat", header = T)
# Marginal parameters, ordered by variable, then parameter
if (!opt$fix.marginals)
    marg.log <- read.table("marginal.mcmc", header = F) 
# If copula is Gaussian, this contains samples for the implied copula
# correlation matrix in the most probable tree    
if (cop.type == "norm")
    implied <- read.table("param.impl.dat", header = T)
    
# Display progress

x11()
acpt.nrow <- nrow(acpt.trees)
plot(1:acpt.nrow, acpt.trees[1:acpt.nrow, 1], "l", col = "red",
     main = "Tree acceptance", ylab = "Acceptance", xlab = "Iteration",
     ylim = c(0, 0.4))
x11()     
min.p  <- min(predictive)
max.p  <- max(predictive)
p.nrow <- length(predictive)
plot(1:p.nrow, predictive, "l", col = "red", main = "Predictive log-likelihood",
     ylab = "Average log-likelihood", xlab = "Iteration",
     ylim = c(min.p - 0.1, max.p + 0.1))

x11()     
num.params <- ncol(marg.log) / ncol(train.x)
par(mfrow = c(1, num.params))
for (i in 1:num.params)
     plot(marg.log[, i], type = "l", xlab = "MCMC iteration", 
          ylab = paste("Parameter #", i, sep = ""))
mtext("Marginal parameters for first variable", outer = TRUE, line = -2)

# Also, one could try this
# > library(coda)
# > plot(as.mcmc(marg.log))
# > effectiveSize(marg.log)
# > plot(as.mcmc(copula.log))

# Provide comparisons

cat("\nSUMMARY OF RESULTS\n")
cat("------------------\n\n")
cat("Predictive log-likelihood (standard structure sampler):", tree.output[1], "\n")
cat("Tree acceptance rate:", tree.output[3], "\n");
cat("---\n")

if (prob.miss > 0) {
    imputed.data <- complete.set.2(train.x, opt$missing.pattern)
    default <- compare.sets(train.x, imputed.data, opt$missing.pattern)
    cat("Default imputation RMSE: ", default, "\n")
    cat("------------------\n\n")
}

cat("Wallclock time:", t.end - t.start, "\n")

################################################################################
## Examples of marginal and copulas
#
# NORMAL MARGINALS:
#
# marg.type   <- "norm"
# marg.priors <- list(list())
# marg.mh     <- list(list(mu.s2q = 0.01, s2.prop = 0.9))
#
# STUDENT T MARGINALS:
#
# marg.type   <- "t"
# marg.priors <- list(list())
# marg.mh     <- list(list(mu.s2q = 0.01, s2.prop = 0.9, nu.prop = 0.9))
#
# GAMMA MARGINALS:
#
# marg.type   <- "gamma"
# marg.priors <- list(list(a.a = 2, a.b = 1, b.a = 2, b.b = 1))
# marg.mh     <- list(list(a.win = 0.95, b.win = 0.95))
#
# EXPONENTIAL MARGINALS:
#
# marg.type   <- "exponential"
# marg.priors <- list(list(a.a = 2, a.b = 1))
# marg.mh     <- list(list(a.win = 0.95))
#
# EMPIRICAL MARGINALS:
#
# marg.type   <- "empirical"
# marg.priors <- list(list())
# marg.mh     <- list(list())
#
#####
#
# NORMAL COPULA
#
# cop.type    <- "norm"
# cop.priors  <- list(list())
# cop.mh      <- list(win = 0.15)
#
# FARLIE-GUMBEL-MORGENSTERN (FGM) COPULA
# 
# cop.type    <- "fgm"
# cop.priors  <- list()
# cop.mh      <- list(win = 0.25)
#
# ALI-MIKHAIL-HAQ (AMH) COPULA
#
# cop.type    <- "amh"
# cop.priors  <- list()
# cop.mh      <- list(win = 0.25)
#
# CLAYTON COPULA
#
# cop.type    <- "clayton"
# cop.priors  <- list(mu = 0, s2 = 10)
# cop.mh      <- list(win = 0.25)
#
# GUMBEL-BARNETT COPULA
#
# cop.type    <- "gb"
# cop.priors  <- list()
# cop.mh      <- list(win = 0.25)
#
# FRANK COPULA
#
# cop.type    <- "frank"
# cop.priors  <- list(mu = 0, s2 = 10)
# cop.mh      <- list(win = 0.25)
