#ifndef __GRAPHCORE_H__
#define __GRAPHCORE_H__ 

extern "C"
{
#include "rhelp.h"
}

#include <stdlib.h>
#include <vector>

using namespace std;

/*
 * CLASS for the generic implementation of graph data structures.
 * For now, I'm focusing on a bare-bones implementation of an
 * undirected graph and a tree graph.
 *
 * All graphs are assumed to be undirected.
 */

class Graph
{
 protected:

  void generateAdjacencyInformation();

 public:


  Graph(int* adjacencies, unsigned int num_nodes);
  ~Graph(void);

  int*  adj_matrix;
  unsigned int num_nodes;
  unsigned int num_edges;

  unsigned int* edge_index;
  vector< vector<unsigned int> > adj_list;
  unsigned int* edges;

  unsigned int getAdjacency(unsigned int pos1, unsigned int pos2);

};

/**
 * TREE DATA STRUCTURE:
 *
 * - order:      partial order. An order that defines a directed tree. 
 *               Nodes later in the order are never parents of nodes earlier
 *               in the order. 
 * - parents:    the parent of each node according to the directed version of 
 *               the tree. Useful for sampling.                 
 * - components: the components of the tree. In the future, we should implement
 *               it for the general graph, and not only the tree case.
 */

class Tree : public Graph
{
 private:

	 void partialOrderRecursion(int current_node, bool *marked, int *level);
     bool isConnected(int *adjacencies, unsigned int num_nodes, int start, int end);
     bool isConnectedRecurse(int *adjacencies, int num_nodes, int start, int end, bool marked[]) ;

 protected:

	 void generateTreeInformation();
     bool findPathRecurse(int start, int end, int path_buffer[], bool marked[]);
     bool findPathRecurseDebug(int start, int end, int path_buffer[], bool marked[], int level);

 public:

  vector< vector<unsigned int> > components;
  unsigned int        num_components;
  unsigned int        *order;
  int                 *parents;

  Tree(int* adjacencies, unsigned int num_nodes);
  ~Tree(void);

  Tree* proposeAddEdge();    Tree* proposeAddEdge(int *end_points); Tree* proposeAddEdge(int *end_points, int *tabu) ;
  Tree* proposeRemoveEdge(); Tree* proposeRemoveEdge(int *end_points);
  Tree* proposeMoveEdge();   Tree* proposeMoveEdge(int *end_points_source, int *end_points_dest);
  Tree* proposeMoveEdgeVertex(int *three_points); Tree* proposeMoveEdgeVertexOld(int *three_points);
  Tree* proposeMoveUniformTreeangle(int *three_points, double *prob_fwd, double *prob_back);
  void  proposeMoveUniformTreeangle(int *three_points) ;
  bool  findPath(int start, int end, int path_buffer[]);
  bool  findPathDebug(int start, int end, int path_buffer[]);
  Tree* clone();
  void  copyAdjacencies(double *out);
  void  printTreeInformation(); void printTreeInformationR();

  bool validAdjacencies(int* adjacencies, unsigned int num_nodes);
};

Tree* sample_unif_tree(int p);    //samples a spanning tree from the uniform distribution.
Tree* sample_chain_tree(int p);   //generates a chain 1 - 2 - 3 - ... - p

extern "C"
{
  void debug_tree(int *matrix, int *num_nodes, int *output);
}

#endif
