extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "marginal.h"
#include "margexp.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
using namespace std;


/*
 * Exponential:
 * 
 * constructor function that sets x to be null
 */

Exponential::Exponential(double a) : Marginal()
{
  this->a = a;
  assert(a > 0);
  this->a_win = 0.01;

  this->a_a = 1;
  this->a_b = 1;
}


/*
 * Exponential:
 * 
 * constructor function that takes new settings of the parameters,
 * but copies data and prior and RW proposal parameters from an old
 * module
 */


Exponential::Exponential(double a, double a_a, double a_b, const Exponential *old)
{
  /* copy data from the old module */
  this->n = old->n;
  this->x = old->x;

  /* copy priors and RW parameters form the old module */
  this->a_win = old->a_win;

  /* copy the new parameter settings */
  assert(a > 0);
  this->a = a;

  this->a_a = a_a; assert(this->a_a > 0);
  this->a_b = a_b; assert(this->a_b > 0);

  /* calculate the ranks */
  CalcRanks();
}


/*
 * Exponential:
 * 
 * constructor function that takes all parameters in the model
 * including the priors and RW proposal parameters through the
 * double vector d
 */

Exponential::Exponential(unsigned n, double *x, double *d) : Marginal(n, x)
{    
  /* copy in the parameters */
  this->a = d[0];
  assert(a > 0);
  this->a_win = d[1];  
  assert(a_win > 0);

  this->a_a = d[2]; assert(this->a_a > 0);
  this->a_b = d[3]; assert(this->a_b > 0);

  /* calculate the ranks */
  CalcRanks();
}


/* 
 * ~Exponential:
 * 
 * destructor
 */

Exponential::~Exponential(void)
{
}


/*
 * NewRanks:
 *
 * populate z with the ranks passed in and calculate
 * the corresponding sample x -- usually called by a
 * copula function
 */

double* Exponential::NewRanks(unsigned int n, double *z)
{
  this->n = n;

  assert(this->z == NULL);
  this->z = z;

  assert(x == NULL);
  x = new_vector(n);
  for(unsigned int i = 0; i < n; i++) 
    x[i] = qexp(z[i], this->a, 1, 0);

  return x;
}


/*
 * CalcRanks:
 *
 * calculate the ranks based on the sample, x,
 * and the parameters mu and s2
 */

void Exponential::CalcRanks(void)
{
  if(z) free(z);
  z = new_vector(n);
  for(unsigned int i = 0; i < n; i++) {
    z[i] = pexp(x[i], this->a, 1, 0);
    if (z[i] < 1.e-16) z[i] = 1.e-16;
    else if (z[i] > 1. - 1.e-16) z[i] = 1. - 1.e-16;
  }
}


/*
 * CalcRanks:
 * 
 * Calculate the rank of a particular data point x
 */

double Exponential::CalcRanks(double x) const
{
  double z = pexp(x, this->a, 1, 0);
    if (z < 1.e-16) z = 1.e-16;
    else if (z > 1. - 1.e-16) z = 1. - 1.e-16;
  return z;
}


/*
 * LLik:
 * 
 * calculate the log likelihood based on the current
 * parameter setting
 */

double Exponential::LLik() const
{
  assert(x);
  double llik = 0;
  for(unsigned int i=0; i < n; i++)
    llik += dexp(x[i], this->a, 1);
  return llik;
}

/*
 * LLik(x):
 * 
 * calculate the log likelihood based on the current
 * parameter setting for a given data point x
 */

double Exponential::LLik(double x) const
{
  return dexp(x, this->a, 1);
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this marginal
 */

unsigned int Exponential::NumParams(void) const
{
  return 1;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void Exponential::Params(double *out) const
{
  out[0] = this->a;
}


/*
 * RW:
 *
 * Random Walk proposals for this marginal -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Marginal* Exponential::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch (which) {
     case 1: return RWa(lqfwd, lqbak);
     default: error("Exponential RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWa:
 *
 * Randowm Walk proposal for a
 */

Marginal* Exponential::RWa(double *lqfwd, double *lqbak) const
{
  //double a_prime = unif_propose_bound(this->a, 0., 10000., this->a_win, lqfwd, lqbak);
  double a_prime = unif_propose_pos(this->a, lqfwd, lqbak, this->a_win);
  return new Exponential(a_prime, this->a_a, this->a_b, this);
}


/*
 * RWvalue:
 *
 * Randowm Walk proposal for the variable at a current value 'value'.
 */

double Exponential::RWvalue(double value, double *lqfwd, double *lqbak) const
{
  //double value_prime = unif_propose_bound(value, 0., 10000., 0.01, lqfwd, lqbak);
  double value_prime = unif_propose_pos(value, lqfwd, lqbak, this->a_win);
  return value_prime;
}


/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 */

double Exponential::LPrior(void) const
{
  return dgamma(this->a, this->a_a, this->a_b, 1);
  return 0.;
}

/*
 * setParams:
 *
 * Sets new parameter values and reset the ranked data
 */

void Exponential::setParams(double *in) 
{
  this->a = in[0];
  assert(this->a > 0);
  CalcRanks();
}

/*
 * randomize:
 *
 * initialize parameters to random numbers.
 * THIS IS NOT SAMPLING FROM THE PRIOR.
 */

void Exponential::randomize()
{    
    this->a = runif(1, 2);
    CalcRanks();
}
