extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "marginal.h"
#include "margnormal.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
#include <R.h>
using namespace std;


/*
 * Normal:
 * 
 * constructor function that sets x to be null
 */

Normal::Normal(double mu, double s2) : Marginal()
{
  this->mu = mu;
  this->mu_s2q = 1;
  this->s2 = s2;
  assert(s2 > 0);
  this->s2_prop = 2.0/3.0;
}

/*
 * Normal:
 * 
 * constructor function that takes new settings of the parameters,
 * but copies data and prior and RW proposal parameters from an old
 * module
 */


Normal::Normal(double mu, double s2, const Normal *old)
{
  /* copy data from the old module */
  this->n = old->n;
  this->x = old->x;

  /* copy priors and RW parameters form the old module */
  this->mu_s2q = old->mu_s2q;
  this->s2_prop = old->s2_prop;

  /* copy the new parameter settings */
  assert(s2 > 0);
  this->mu = mu;
  this->s2 = s2;

  /* calculate the ranks */
  CalcRanks();
}


/*
 * Normal:
 * 
 * constructor function that takes all parameters in the model
 * including the priors and RW proposal parameters through the
 * double vector d.
 */

Normal::Normal(unsigned n, double *x, double *d) : Marginal(n, x)
{
  /* copy in the parameters */
  this->mu = d[0];
  this->s2 = d[1];
  assert(s2 > 0);
  this->mu_s2q = d[2];
  assert(mu_s2q > 0);
  this->s2_prop = d[3];
  assert(s2_prop > 0 && s2_prop < 1);

  /* calculate the ranks */
  CalcRanks();
}


/* 
 * ~Normal:
 * 
 * destructor
 */

Normal::~Normal(void)
{
}


/*
 * NewRanks:
 *
 * populate z with the ranks passed in and calculate
 * the corresponding sample x -- usually called by a
 * copula function
 */

double* Normal::NewRanks(unsigned int n, double *z)
{
  this->n = n;

  assert(this->z == NULL);
  this->z = z;

  assert(x == NULL);
  x = new_vector(n);
  double ss2 = sqrt(s2);
  for(unsigned int i = 0; i < n; i++) {
    x[i] = qnorm(z[i], mu, ss2, 1, 0);
  }
  return x;
}


/*
 * CalcRanks:
 *
 * calculate the ranks based on the sample, x,
 * and the parameters mu and s2
 */

void Normal::CalcRanks(void)
{
  if (this->z) free(this->z);
  this->z = new_vector(this->n);
  double ss2 = sqrt(this->s2);
  for (unsigned int i = 0; i < this->n; i++) {
      this->z[i] = pnorm(this->x[i], this->mu,  ss2, 1, 0);
      if (this->z[i] == 0.) this->z[i] = 1.e-16;
      else if (this->z[i] == 1.) this->z[i] = 1. - 1.e-16;
  }
}


/*
 * CalcRanks:
 *
 * calculate the ranks for a particular point, x
 * under this marginal
 */

double Normal::CalcRanks(double x) const
{
  double ss2 = sqrt(this->s2);
  double z = pnorm(x, this->mu, ss2, 1, 0);
    if (z == 0.) z = 1.e-16;
    else if (z == 1.) z = 1. - 1.e-16;
  return z;
}


/*
 * LLik:
 * 
 * returns the log likelihood based on the current
 * parameter setting
 */

double Normal::LLik() const
{
  assert(this->x);
  double llik = 0;
  double ss2 = sqrt(this->s2);
  for (unsigned int i = 0; i < this->n; i++) {
      //Rprintf("LLIK(%d; %f) = %f\n", i, x[i], dnorm(this->x[i], this->mu, ss2, 1));
      llik += dnorm(this->x[i], this->mu, ss2, 1);
  }
  return llik;
}

/*
 * LLik(x):
 * 
 * calculate the log likelihood based on the current
 * parameter setting for a given data point x
 */

double Normal::LLik(double x) const
{
  double ss2 = sqrt(this->s2);
  return dnorm(x, this->mu, ss2, 1);
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this marginal
 */

unsigned int Normal::NumParams(void) const
{
  return 2;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void Normal::Params(double *out) const
{
  out[0] = mu;
  out[1] = s2;
}


/*
 * RW:
 *
 * Random Walk proposals for this marginal -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Marginal* Normal::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWmu(lqfwd, lqbak);
    case 2: return RWs2(lqfwd, lqbak);
    default: error("Normal RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWmu:
 *
 * Randowm Walk proposal for the mean
 */

Marginal* Normal::RWmu(double *lqfwd, double *lqbak) const
{
  double ss2q = sqrt(this->mu_s2q);
  double mu_prime = rnorm(this->mu, ss2q);
  //Rprintf("Proposing mu %f --> %f", this->mu, mu_prime);
  *lqfwd = *lqbak = dnorm(this->mu, mu_prime, ss2q, 1);
  return new Normal(mu_prime, this->s2, this);
}


/*
 * RWs2:
 *
 * Randowm Walk proposal for the variance
 */ 

Marginal* Normal::RWs2(double *lqfwd, double *lqbak) const
{
  double s2_prime = unif_propose_pos(this->s2, lqfwd, lqbak, this->s2_prop);
  //Rprintf("Proposing s2 %f --> %f\n", this->s2, s2_prime);
  return new Normal(this->mu, s2_prime, this);
}

/*
 * RWvalue:
 *
 * Randowm Walk proposal for the variable at a current value 'value'.
 */

double Normal::RWvalue(double value, double *lqfwd, double *lqbak) const
{
  double ss2q = sqrt(mu_s2q);
  double value_prime = rnorm(value, ss2q);
  *lqfwd = *lqbak = dnorm(value, value_prime, ss2q, 1);
  return value_prime;
}

/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters.
 * Improper prior in this case.
 */

double Normal::LPrior(void) const
{
  return 0.0 - log(this->s2); 
}

/*
 * setParams:
 *
 * Sets new parameter values and reset the ranked data
 */

void Normal::setParams(double *in) 
{
  this->mu = in[0];
  this->s2 = in[1];
  assert(this->s2 > 0);
  CalcRanks();
}

/*
 * randomize:
 *
 * initialize parameters to random numbers.
 * THIS IS NOT SAMPLING FROM THE PRIOR.
 */

void Normal::randomize()
{
  this->mu = runif(-1, 1);
  this->s2 = runif(0.1, 2);
  CalcRanks();
}
