#ifndef __STRUCTURE_H__
#define __STRUCTURE_H__ 

#include "graphcore.h"
#include "copula.h"
#include <set>
using namespace std;


class MCMCStructure
{
private:

  unsigned int quadrature_size;
  unsigned int total_params;
  double **copula_quadrature;

  int *initial_clustering;

  double *uv, *u_s, *u_sq, *del_score, *add_score;
  double **u;
  unsigned int *edge_buffer;  

  double *thetaSpace;

  unsigned int randomSuccessor(unsigned int u, double *weights);  
  double *startCopulaQuadrature(unsigned int u, unsigned int v);

  void inverse(double *M, int m);

protected:

  bool    verbose;
  bool    gaussian_copula, one_param;
  int     tree_prop;
  bool    do_marg_update;
  bool    do_indcopula_update;

  ofstream z_file; ifstream z_in; 
  ofstream tree_file; ifstream tree_in;
  ofstream k_file;  ifstream k_in;
  ofstream param_file;  ifstream param_in;
  ofstream marginal_file;
  ofstream accepttree_file;
  ofstream sel_copula_file;
  ofstream impl_gauss_file;

  int     accepted_tree, accepted_parameter;  
  unsigned int M;          /* number of extra components for Neal's ``Algorithm 8''*/

  bool    useCopulaSlice;   /* If true, allows the use of the slice sampler for copula parameters */

  unsigned int num_missing_rows, *num_missing_cols;       /* Number of row/columns with missing data. */
  unsigned int *missing_rows;      /* Index of the rows with missing data. */
  unsigned int **missing_columns;  /* Index of the columns with missing data (corresponding to the missing rows). */

  Tree **initial_trees;      /* Trees that initialize the Markov chain */
  double *initial_params;    /* Parameters that initialize the Markov chain */

  double *marginal_data;   /* copy of the original data, but ordered columnwise */
  double *imputedData;     /* buffer that stores the average missing data predictions */
  double *dataSnapshot;    /* buffer that stores a copy of the original dataset before being modified by augmentation */
 
  Marginal **all_M;        /* marginals: one for each random variable */
  Copula   **all_cop;      /* copulas: one for each pair of random variables */

  double *md, *copd;
  unsigned int md_size, copd_size; /* Size of each marginal and copula specification. We assume they
				      are ALL OF THE SAME SIZE, even if it means some marginal
				      (or copula) specifications will be padded with zeroes. */
  int    *z, *z_count;             /* Indicator of model component in the mixture of trees. */
  double alpha, a, b;              /* Prior and hyperprior parameters for the Dirichlet process.
                                      Here, alpha follows a Gamma(a, b).*/

  unsigned int K, initial_K, nonzero_K; /* Number of components */
  Tree     **tree_mix;                  /* Trees of each component */
  Copula   **cop_mix;                   /* Copulas of each component */

  set<int> **idx;                       /* idx[k] lists the set of data points associated with tree_mix[k] */

  bool random_tree; /* if true, generate initialize the chain at a random tree. */

  /* Pred-loglikelihood + test set data structures */
  unsigned int burn_in, n_test, mcmc_step;
  double *test_data;
  int    *mars_matrix_test;
  double *predictions;

  void initializeMissingDataInfo();

  double pointLogLikelihood(unsigned int d, unsigned int k);
  void getMode(int *mode, unsigned int num_samples, double *samples);

  void completeData(double *data, unsigned int n_d, int *mars_matrix_data);

  void addClusters(unsigned int new_K, unsigned int *data_points);
  void printClusters();
  void packClusters();
  void indexEdges();

  void updateAlpha();
  void updateZ();  
  void resampleParameters(unsigned int k);
  void updateStructure(unsigned int k);
  void updateStructureRandomized(unsigned int k);  void updateStructureRandomWalk(unsigned int k); 
  void updateStructureStandard(unsigned int k); void updateStructureOneParam(unsigned int k); 
  void updateStructureOneParamHybrid(unsigned int k);
  void updateStructureExact(unsigned int k);
  void computeComponentProbability(int i, double *p);
  void copy_from_old(unsigned int k, unsigned int pos, int *old_z_count, 
                     Tree **old_tree_mix, Copula **old_cop_mix);
  void updateMarginals(unsigned int v); 
  void updateCopulas(unsigned int k, unsigned int v1, unsigned int v2);
       void updateCopulasMH(unsigned int k, unsigned int v1, unsigned int v2); 
       void updateCopulasSlice(unsigned int k, unsigned int v1, unsigned int v2);
  void updateMissingData(); void addMissingDataPrediction();

  double copulaLogLikelihood(Tree *tree, unsigned int k, double *point);
  double marginalsLogLikelihood(double *point);

  void WriteZFile();
  void WriteKFile();
  void WriteTreeFile(double *tree_buffer);
  void WriteParamsFile(); void WriteMarginalFile();
  void WriteAcceptedTreeFile(int possible_trees);
  void WriteSelectedCopulaFile();
  void WriteImpliedGaussianParameterSample();
  
  void RoundsInit(unsigned int T, unsigned int predict_out, unsigned int track_missing);
  void step();
  void RoundsCleanup(unsigned int predict_out, unsigned int track_missing);

  void   PredInit(unsigned int burn_in);
  void   GetTreesIn(void);
  void   GetParamsIn(void);
  void   predLikelihood(double *pred_llik, double *test, unsigned int n_t, int *z);
  double predLoglikelihood(unsigned int i, double *test, int *z);

  void   storePredictiveLogLikelihood(unsigned int t);
  double calculatePredictiveLoglikelihood(unsigned int T);

  double calcPostEdgeWeight(unsigned int k, unsigned int u, unsigned int v);

public:

  unsigned int n; //Sample size of sample (used in maximum likelihood estimation and posterior inference)
  unsigned int p; //Dimensionality of the data (number of variables)

  double       *data;
  int          *mars_matrix; //matrix for missing at random indicators: 1 for recorded, 0 for hidden

  MCMCStructure(unsigned int n, unsigned int p, double a, double b, 
                    double *data, int *mars_matrix,
                    double *md, unsigned int md_size,
                    double *copd, unsigned int copd_size,
                    unsigned int initial_K, int *initial_clustering, Tree **initial_trees, double *initial_params);
  ~MCMCStructure();

  void   Rounds(unsigned int T, unsigned int predict_out, unsigned int track_missing, double output_buffer[]);   

  double evaluateMissingValues();
  double predLogLikelihood(unsigned int T, unsigned int burn_in, double *test, unsigned int n_t, int *mars_matrix_test);

  void setBurnIn(unsigned int burn_in);
  void setMCMCStep(unsigned int mcmc_step);
  void setTestSet(unsigned int n_test, double *test_data, int *mars_matrix_test);
  void setTreeProposal(int v);
  void setMarginalUpdater(bool m);
  void setCopulaSliceSampler(bool u);
  void setIndCopulaUpdate(bool u);
};

Marginal* new_marg(unsigned int n, double *x, double *d);
Copula*   new_copula(double *d);

extern "C" {
    void mcmc_structure_treevariate(int *n, int *p, double *a, double *b, 
                                    double *data, int *mars_matrix,
                                    int *n_test, double *test_data, int *mars_matrix_test,
                                    double *md, int *md_size, double *copd, int *copd_size,
                                    int    *num_samples, int *burn_in, int *mcmc_step, 
                                    int *initial_K, int *initial_clustering, int *initial_adj, double *initial_params,
                                    int *options, double *output);
}

#endif
